#ifndef APP_H
#define APP_H

#include "DXUTgui.h"
#include "DXUTmisc.h"
#include "DXUTCamera.h"
#include "DXUTSettingsDlg.h"
#include "SDKmisc.h"
#include "SDKmesh.h"

// App includes
#include <vector>
#include <string>
#include <stack>

#include "Camera.h"
#include "LightMaterial.h"
#include "Utilities.h"
#include "Vertex.h"

//#define DEBUG_VS   // Uncomment this line to debug D3D9 vertex shaders 
//#define DEBUG_PS   // Uncomment this line to debug D3D9 pixel shaders 

#define ZNEARPLANE		0.1f
#define ZFARPLANE		10000.0f
#define FOV				D3DX_PI * 0.25f

//--------------------------------------------------------------------------------------
// UI control IDs
//--------------------------------------------------------------------------------------
enum { IDC_TOGGLEFULLSCREEN, IDC_TOGGLEREF, IDC_CHANGEDEVICE, 
	IDC_TOGGLEWIREFRAME, IDC_SHOWGUI, IDC_CAMSPEEDTEXT, IDC_SLIDERCAMSPEED, 
	IDC_CLIENTCONTROLID };

class App; 

// Function to initialize DXUT and application
HRESULT Initialize( int width, int height, const WCHAR *pWindowTitle, App *app );

// Main application framework
class App
{
public:
	App(void);
	virtual ~App(void);	

	void OneTimeInit(); 			
	void RenderText();	
	
	// App callback functions
	HRESULT CALLBACK OnCreateDevice( IDirect3DDevice9* pd3dDevice, const D3DSURFACE_DESC* pBackBufferSurfaceDesc );		
	virtual HRESULT OnResetDevice();
	virtual void OnDestroyDevice();
	virtual void OnLostDevice();			
	void OnFrameRender( double fTime, float fElapsedTime );
	void OnFrameMove( double fTime, float fElapsedTime );	
	LRESULT MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam );
	void KeyboardProc( UINT nChar, bool bKeyDown, bool bAltDown );	
	void MouseProc( bool bLeftButtonDown, bool bRightButtonDown, bool bMiddleButtonDown, bool bSideButton1Down, bool bSideButton2Down, int nMouseWheelDelta, int xPos, int yPos );
	void GUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl );	

	// Client callback functions
	virtual HRESULT OneTimeClientInit()													{ return S_OK; }
	virtual HRESULT ResetClient()														{ return S_OK; }
	virtual void LostClient()															{ }
	virtual void DestroyClient()														{ }
	virtual void RenderClient( double time, float elapsedTime )							{ }
	virtual void UpdateClient( double time, float elapsedTime )							{ }	
	virtual void GUIEventClient( UINT nEvent, int nControlID, CDXUTControl* pControl )	{ }
	virtual void KeyboardProcClient( UINT nChar, bool bKeyDown, bool bAltDown )			{ }
	virtual LRESULT MsgProcClient( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )	{ return 0; }
	virtual void MouseProcClient( bool bLeftButtonDown, bool bRightButtonDown, bool bMiddleButtonDown, bool bSideButton1Down, bool bSideButton2Down, int nMouseWheelDelta, int xPos, int yPos ) { }

public:
	// Accessors
	CD3DSettingsDlg *			GetD3DSettingsDialog()	{ return &mSettingsDlg; }
	CDXUTDialog		*			GetAppDlg()				{ return &mAppDlg; }

	bool IsWireFrame()			{ return mAppDlg.GetCheckBox( IDC_TOGGLEWIREFRAME )->GetChecked(); }	
	bool IsAppExiting()			{ return mAppExiting; }
	void ConfineMouse();	
	
protected:
	void RenderAxis();
	void RenderGrid();	
	void CreateGridPoints(int numRows, int numCols, float unitSize);	
	D3DXVECTOR3 GetRayTo(int x, int y);

protected:
	// Helper funtions to add to stats	
	void AddTriangles( DWORD c )		{ mTotalTriangles += c; }
	void AddVertices( DWORD c )			{ mTotalVertices += c; }	
	void AddDebugStr( char str[512] )	{ mDebugStrings.push_back( str ); }		
		
	std::vector<std::string>	mDebugStrings;

private:	
	// Some variables for stats	
	DWORD mTotalTriangles;
	DWORD mTotalVertices;		

	// Other fps stats	
	float mAvgFps;
	float mBestFps;
	float mWorstFps;	

protected:
	LPDIRECT3DDEVICE9			mDevice;				// Store the pointer to the Direct3D device

	CDXUTDialogResourceManager  mDialogResourceManager;	// Manager for shared resources of dialogs
	CD3DSettingsDlg             mSettingsDlg;			// Device settings dialog
	CDXUTDialog                 mAppDlg;				// Dialog for app specific controls	
	LPD3DXFONT                  mFont;					// Needed for text helper	
	LPD3DXSPRITE                mSprite;				// Needed to draw 2d textures and text
	LPD3DXLINE					mLine;					// For any line drawing requirements
	
	HWND						mWndHandle;				// App window handle
	HINSTANCE					mInstHandle;			// Application instance handle
	int							mScreenWidth;			// Application screen width
	int							mScreenHeight;			// Applicatoin screen height
	Camera						mCamera;				// Free roam camera	
	Light						mDirLight;				// Scene directional light
	
	char						mDebugString[512];		// Useful string container
	bool						mRightButtonDown;

	int							mCurrentGuiHeight;
	float						mCamSpeed;	
	bool						mAppExiting;	
	bool						mShowGui;

	// 3D grid variables
	D3DXVECTOR3  			*mGridPoints;
	DWORD					mNumGridPrimitives;	
};

// Globals for convenient access.
static App* gApp = NULL;

//--------------------------------------------------------------------------------------
// DXUT callback functions forward declarations 
//--------------------------------------------------------------------------------------
LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool* pbNoFurtherProcessing, void* pUserContext );
void CALLBACK OnKeyboard( UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext );
void CALLBACK OnMouseProc( bool bLeftButtonDown, bool bRightButtonDown, bool bMiddleButtonDown, bool bSideButton1Down, bool bSideButton2Down, int nMouseWheelDelta, int xPos, int yPos, void* pUserContext );
void CALLBACK OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext );
void CALLBACK OnFrameMove( double fTime, float fElapsedTime, void* pUserContext );
bool CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, void* pUserContext );
bool CALLBACK IsD3D9DeviceAcceptable( D3DCAPS9* pCaps, D3DFORMAT AdapterFormat, D3DFORMAT BackBufferFormat, bool bWindowed, void* pUserContext );
HRESULT CALLBACK OnD3D9CreateDevice( IDirect3DDevice9* pd3dDevice, const D3DSURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext );
HRESULT CALLBACK OnD3D9ResetDevice( IDirect3DDevice9* pd3dDevice, const D3DSURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext );
void CALLBACK OnD3D9FrameRender( IDirect3DDevice9* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext );
void CALLBACK OnD3D9LostDevice( void* pUserContext );
void CALLBACK OnD3D9DestroyDevice( void* pUserContext );

#endif