#include "dxut.h"
#include "Camera.h"

Camera::Camera(void)
{
	LOG_DEBUG( "Camera constructor called." );

	m_vMinBoundary = D3DXVECTOR3( -175.0f, 2.5f, -175.f );
	m_vMaxBoundary = D3DXVECTOR3( 175, 50.f, 175.f );
	m_bClipToBoundary = true;

	m_fCameraPitchAngle = m_fCameraYawAngle = 0.5f;
}

Camera::~Camera(void)
{
	LOG_DEBUG( "Camera destructor called." );
}

void Camera::FrameMove( float fElapsedTime )
{
	POINT pt;
	GetCursorPos( &pt );
	ScreenToClient( DXUTGetHWND(), &pt );

	if( DXUTGetGlobalTimer()->IsStopped() )
		fElapsedTime = 1.0f / DXUTGetFPS();

	if( IsKeyDown( m_aKeys[CAM_RESET] ) )
		Reset();

	// Get keyboard/mouse/gamepad input
	GetInput( m_bEnablePositionMovement, ( m_nActiveButtonMask & m_nCurrentButtonMask ) || m_bRotateWithoutButtonDown, true, m_bResetCursorAfterMove );

	// Get amount of velocity based on the keyboard input and drag (if any)
	UpdateVelocity( fElapsedTime );

	// Simple euler method to calculate position delta
	D3DXVECTOR3 vPosDelta = m_vVelocity * fElapsedTime;

	// If rotating the camera 
	if( ( m_nActiveButtonMask & m_nCurrentButtonMask ) ||
		m_bRotateWithoutButtonDown ||
		m_vGamePadRightThumb.x != 0 ||
		m_vGamePadRightThumb.z != 0)
	{
		// Update the pitch & yaw angle based on mouse movement
		float fYawDelta = m_vRotVelocity.x;
		float fPitchDelta = m_vRotVelocity.y;

		// Invert pitch if requested
		if( m_bInvertPitch )
			fPitchDelta = -fPitchDelta;

		m_fCameraPitchAngle += fPitchDelta;
		m_fCameraYawAngle += fYawDelta;
	}

	// Limit pitch angle
	if(m_fCameraPitchAngle > (D3DX_PI / 2.0f) - 0.05f) 
	{	
		m_fCameraPitchAngle = D3DX_PI / 2.0f - 0.05f;
	}
	if(m_fCameraPitchAngle < 0.3f)
	{
		m_fCameraPitchAngle = 0.3f;
	}

	// Make a rotation matrix based on the camera's yaw & pitch
	D3DXMATRIX mCameraRot;
	D3DXMatrixRotationYawPitchRoll( &mCameraRot, m_fCameraYawAngle, m_fCameraPitchAngle, 0 );

	// Transform vectors based on camera's rotation matrix
	D3DXVECTOR3 vWorldUp, vWorldAhead;
	D3DXVECTOR3 vLocalUp = D3DXVECTOR3( 0, 1, 0 );
	D3DXVECTOR3 vLocalAhead = D3DXVECTOR3( 0, 0, 1 );
	D3DXVec3TransformCoord( &vWorldUp, &vLocalUp, &mCameraRot );
	D3DXVec3TransformCoord( &vWorldAhead, &vLocalAhead, &mCameraRot );

	// Transform the position delta by the camera's rotation 
	D3DXVECTOR3 vPosDeltaWorld;
	if( !m_bEnableYAxisMovement )
	{
		// If restricting Y movement, do not include pitch
		// when transforming position delta vector.
		D3DXMatrixRotationYawPitchRoll( &mCameraRot, m_fCameraYawAngle, 0.0f, 0.0f );
	}
	D3DXVec3TransformCoord( &vPosDeltaWorld, &vPosDelta, &mCameraRot );

	// Move the eye position 
	m_vEye += vPosDeltaWorld;

	if( m_bClipToBoundary )
		ConstrainToBoundary( &m_vEye );

	// Update the lookAt position based on the eye position 
	m_vLookAt = m_vEye + vWorldAhead;

	D3DXMatrixLookAtLH( &m_mView, &m_vEye, &m_vLookAt, &D3DXVECTOR3( 0, 1, 0 ) );

	D3DXMatrixInverse( &m_mCameraWorld, NULL, &m_mView );
	
	UpdateFrustum();
}

void Camera::UpdateFrustum()
{
	// Get combined matrix
	D3DXMATRIX matComb;
	D3DXMatrixMultiply(&matComb, &m_mView, &m_mProj);

	// Left clipping plane
	mPlanes[0].a = matComb._14 + matComb._11; 
	mPlanes[0].b = matComb._24 + matComb._21; 
	mPlanes[0].c = matComb._34 + matComb._31; 
	mPlanes[0].d = matComb._44 + matComb._41;

	// Right clipping plane 
	mPlanes[1].a = matComb._14 - matComb._11; 
	mPlanes[1].b = matComb._24 - matComb._21; 
	mPlanes[1].c = matComb._34 - matComb._31; 
	mPlanes[1].d = matComb._44 - matComb._41;

	// Top clipping plane 
	mPlanes[2].a = matComb._14 - matComb._12; 
	mPlanes[2].b = matComb._24 - matComb._22; 
	mPlanes[2].c = matComb._34 - matComb._32; 
	mPlanes[2].d = matComb._44 - matComb._42;

	// Bottom clipping plane 
	mPlanes[3].a = matComb._14 + matComb._12; 
	mPlanes[3].b = matComb._24 + matComb._22; 
	mPlanes[3].c = matComb._34 + matComb._32; 
	mPlanes[3].d = matComb._44 + matComb._42;	  

	// Near clipping plane 
	mPlanes[4].a = matComb._14 + matComb._13; 
	mPlanes[4].b = matComb._24 + matComb._23; 
	mPlanes[4].c = matComb._34 + matComb._33; 
	mPlanes[4].d = matComb._44 + matComb._43;

	// Far clipping plane 
	mPlanes[5].a = matComb._14 - matComb._13; 
	mPlanes[5].b = matComb._24 - matComb._23; 
	mPlanes[5].c = matComb._34 - matComb._33; 
	mPlanes[5].d = matComb._44 - matComb._43; 

	//Normalize planes
	for(int i=0;i<6;i++)
		D3DXPlaneNormalize(&mPlanes[i], &mPlanes[i]);
}

bool Camera::FrustumSphereCheck(D3DXVECTOR3 &vCenter, float fRadius)
{
	// Make sure radius is in frustum
	for(int i=0;i<6;i++) 
	{
		if(D3DXPlaneDotCoord(&mPlanes[i], &vCenter) < -fRadius)
			return false;
	}
	return true;
}

bool Camera::FrustumCubeCheck( D3DXVECTOR3 &vCenter, float fSize )
{
	short i;

	// Make sure at least one point is completely in frustum
	for( i = 0; i < 6; i++ )
	{
		if(D3DXPlaneDotCoord(&mPlanes[i], &D3DXVECTOR3(vCenter.x-fSize, vCenter.y-fSize, vCenter.z-fSize)) >= 0.0f)
			continue;
		if(D3DXPlaneDotCoord(&mPlanes[i], &D3DXVECTOR3(vCenter.x+fSize, vCenter.y-fSize, vCenter.z-fSize)) >= 0.0f)
			continue;
		if(D3DXPlaneDotCoord(&mPlanes[i], &D3DXVECTOR3(vCenter.x-fSize, vCenter.y+fSize, vCenter.z-fSize)) >= 0.0f)
			continue;
		if(D3DXPlaneDotCoord(&mPlanes[i], &D3DXVECTOR3(vCenter.x+fSize, vCenter.y+fSize, vCenter.z-fSize)) >= 0.0f)
			continue;
		if(D3DXPlaneDotCoord(&mPlanes[i], &D3DXVECTOR3(vCenter.x-fSize, vCenter.y-fSize, vCenter.z+fSize)) >= 0.0f)
			continue;
		if(D3DXPlaneDotCoord(&mPlanes[i], &D3DXVECTOR3(vCenter.x+fSize, vCenter.y-fSize, vCenter.z+fSize)) >= 0.0f)
			continue;
		if(D3DXPlaneDotCoord(&mPlanes[i], &D3DXVECTOR3(vCenter.x-fSize, vCenter.y+fSize, vCenter.z+fSize)) >= 0.0f)
			continue;
		if(D3DXPlaneDotCoord(&mPlanes[i], &D3DXVECTOR3(vCenter.x+fSize, vCenter.y+fSize, vCenter.z+fSize)) >= 0.0f)
			continue;

		return false;
	}
	return true;
}

bool Camera::FrustumCubeCheck( D3DXVECTOR3 &vMin, D3DXVECTOR3 &vMax )
{
	for( int i = 0; i < 6; i++ )
	{
		if( D3DXPlaneDotCoord( &mPlanes[i], &D3DXVECTOR3( vMin.x, vMin.y, vMin.z ) ) >= 0.0f )
			continue;
		if( D3DXPlaneDotCoord( &mPlanes[i], &D3DXVECTOR3( vMax.x, vMin.y, vMin.z ) ) >= 0.0f )
			continue;
		if( D3DXPlaneDotCoord( &mPlanes[i], &D3DXVECTOR3( vMin.x, vMax.y, vMin.z ) ) >= 0.0f )
			continue;
		if( D3DXPlaneDotCoord( &mPlanes[i], &D3DXVECTOR3( vMax.x, vMax.y, vMin.z ) ) >= 0.0f )
			continue;
		if( D3DXPlaneDotCoord( &mPlanes[i], &D3DXVECTOR3( vMin.x, vMin.y, vMax.z ) ) >= 0.0f )
			continue;
		if( D3DXPlaneDotCoord( &mPlanes[i], &D3DXVECTOR3( vMax.x, vMin.y, vMax.z ) ) >= 0.0f )
			continue;
		if( D3DXPlaneDotCoord( &mPlanes[i], &D3DXVECTOR3( vMin.x, vMax.y, vMax.z ) ) >= 0.0f )
			continue;
		if( D3DXPlaneDotCoord( &mPlanes[i], &D3DXVECTOR3( vMax.x, vMax.y, vMax.z ) ) >= 0.0f )
			continue;

		return false;
	}
	return true;
}