#include "DXUT.h"
#include "Terrain.h"
#include "Vertex.h"
#include "HeightMap.h"
#include "Ray.h"
#include "Camera.h"
#include "Utilities.h"	

TerrainPatch::TerrainPatch()
{
	mDevice = NULL;
	mMesh = NULL;	

}
TerrainPatch::~TerrainPatch()
{
	Release();	
}

void TerrainPatch::Release()
{
	SAFE_RELEASE( mMesh );
}

HRESULT TerrainPatch::CreateMesh(Terrain *pTerrain, RECT source, int index)
{
	HRESULT hr = S_OK;

	// Store a Pointer copy of the d3d device too
	mDevice = pTerrain->mDevice;
	mMapRect = source;

	int width = source.right - source.left;
	int length = source.bottom - source.top;
	int nrVert = (width + 1) * (length + 1);
	int nrTri = width * length * 2;
	Point size = pTerrain->mHeightMap->GetSize();
	float xOffset = (float)size.x * 0.5f; 
	float zOffset = (float)size.y * 0.5f;	

	V(D3DXCreateMeshFVF(nrTri, nrVert, D3DXMESH_DYNAMIC, D3DFVF_VERTEXTERRAIN, mDevice, &mMesh));

	//Create vertices
	VertexTerrain* ver = 0;
	//Vertex * ver = 0;
	mMesh->LockVertexBuffer(0,(void**)&ver);			

	for(int z = source.top, z0 = 0; z <= source.bottom; z++, z0++)			
	{
		for(int x = source.left, x0 = 0; x <= source.right; x++, x0++)	
		{
			D3DCOLOR col;

			if(index % 2 == 0)		//Invert color depending on what patch it is...
				col = D3DCOLOR_ARGB(255, 0, 0, 255);
			else
				col = D3DCOLOR_ARGB(255, 255, 0, 0);

			//Extract height (and position) from heightmap
			float height = pTerrain->mHeightMap->GetHeightTable()[x + z * size.x];
			D3DXVECTOR3 pos = D3DXVECTOR3( x - xOffset, height, z - zOffset);								
			D3DXVECTOR2 alphaUV = D3DXVECTOR2(x / (float)size.x, z / (float)size.y);	// Alpha UV
			D3DXVECTOR2 colorUV = alphaUV * 8.0f;										// Color UV			

			// Set new vertex			
			ver[z0 * (width + 1) + x0].position = pos;
			ver[z0 * (width + 1) + x0].tex0 = alphaUV;
			ver[z0 * (width + 1) + x0].tex1 = colorUV;
			ver[z0 * (width + 1) + x0].color = col;			

			// Calculate bounding box for this patch
			if(pos.x < mBoundingBox.min.x) mBoundingBox.min.x = pos.x;
			if(pos.x > mBoundingBox.max.x) mBoundingBox.max.x = pos.x;
			if(pos.y < mBoundingBox.min.y) mBoundingBox.min.y = pos.y;
			if(pos.y > mBoundingBox.max.y) mBoundingBox.max.y = pos.y;
			if(pos.z < mBoundingBox.min.z) mBoundingBox.min.z = pos.z;
			if(pos.z > mBoundingBox.max.z) mBoundingBox.max.z = pos.z;
		}
		int x = 0;
	}
	mMesh->UnlockVertexBuffer();

	//Calculate Indices
	WORD* ind = 0;
	mMesh->LockIndexBuffer(0,(void**)&ind);	
	int i = 0;

	for( int z = source.top, z0 = 0; z < source.bottom; z++, z0++ )
	{
		for( int x = source.left, x0 = 0; x < source.right; x++, x0++ )
		{
			// Triangle 1
			ind[i++] =   z0   * (width + 1) + x0;			
			ind[i++] = (z0+1) * (width + 1) + x0;		
			ind[i++] =   z0   * (width + 1) + x0 + 1;

			// Triangle 2
			ind[i++] = (z0+1) * (width + 1) + x0;			
			ind[i++] = (z0+1) * (width + 1) + x0 + 1;
			ind[i++] =   z0   * (width + 1) + x0 + 1;
		}
	}

	mMesh->UnlockIndexBuffer();

	// No need to set attributes
	DWORD *att = 0, a = 0;
	mMesh->LockAttributeBuffer(0,&att);
	memset(att, 0, sizeof(DWORD)*nrTri);
	mMesh->UnlockAttributeBuffer();

	//Compute normals
	D3DXComputeNormals(mMesh, NULL);

	return hr;
}

void TerrainPatch::Render(int texture)
{
	// Draw mesh
	if(mMesh != NULL)
	{
		mMesh->DrawSubset(texture);
	}
}

Terrain::Terrain(void)
{
	LOG_DEBUG( " Terrain constructor Called." );

	mEffect		= NULL;
	mHeightMap	= NULL;	
	mLightMap	= NULL;
	mCalculatedBlendMap	= NULL;
}

Terrain::~Terrain(void)
{
	LOG_DEBUG( "Terrain destructor Called." );	
	Release();
}

void Terrain::Reset(LPDIRECT3DDEVICE9 device, HeightMap *heightmap, D3DXVECTOR3 &dirToLight, int numPatches)
{
	mDevice		= device;
	mHeightMap	= heightmap;
	mDirToLight = -dirToLight;	
	
	// Load textures
	LPDIRECT3DTEXTURE9 diffuseMaps[9];

	char *textures[] = { "media/terrain/splatting0.png", 
		"media/terrain/splatting1.png",
		"media/terrain/splatting2.png",
		"media/terrain/splatting3.png",
		"media/terrain/splatting4.png",
		"media/terrain/splatting5.png",
		"media/terrain/splatting6.png",
		"media/terrain/splatting7.png",
		"media/terrain/splatting8.png" };

	// Load diffuse maps
	for( int i = 0; i < 9; i++ )
	{
		if(FAILED(D3DXCreateTextureFromFileA(mDevice, textures[i] , &diffuseMaps[i]))) 
		{ 
			LOG_ERROR("Failed to load terrain texture: " << textures[i]);
		}
		mDiffuseMaps.push_back(diffuseMaps[i]);
	}

	LPDIRECT3DTEXTURE9 blendMap = NULL;
	if(FAILED(D3DXCreateTextureFromFileA(mDevice, "media/terrain/blendmap1.png" , &blendMap))) 
	{ 
		LOG_ERROR("Failed to load terrain blendmap." );
	}
	mBlendMaps.push_back( blendMap );	

	LPDIRECT3DTEXTURE9 blendMap2 = NULL;
	if(FAILED(D3DXCreateTextureFromFileA(mDevice, "media/terrain/blendmap2.png" , &blendMap2))) 
	{ 
		LOG_ERROR("Failed to load terrain blendmap." );
	}
	mBlendMaps.push_back( blendMap2 );		

	LPDIRECT3DTEXTURE9 blendMap3 = NULL;
	if(FAILED(D3DXCreateTextureFromFileA(mDevice, "media/terrain/blendmap3.png" , &blendMap3))) 
	{ 
		LOG_ERROR("Failed to load terrain blendmap." );
	}
	mBlendMaps.push_back( blendMap3 );		

	// Load shader
	ID3DXBuffer* errors = 0;
	if(FAILED(D3DXCreateEffectFromFile( mDevice, L"Terrain.fx",
		0, 0, D3DXSHADER_DEBUG, 0, &mEffect, &errors)))
	{
		LOG_ERROR("Failed to load terrain effect file. " << (char*)errors->GetBufferPointer());	
	}

	CreatePatches(numPatches);	
	CalculateBlendMap();	
	CalculateLightMap();	
}

void Terrain::CreatePatches(int numPatches)
{	
	try
	{
		// Clear any old mPatches
		for( size_t i = 0; i < mPatches.size(); i++ )
		{
			if(mPatches[i] != NULL)
			{
				mPatches[i]->Release();
			}
		}
		mPatches.clear();

		// Create the patches		
		for( int y = 0; y < numPatches; y++ )
		{
			Progress( "Creating patches", y / (float)numPatches );
			for( int x = 0; x < numPatches; x++ )
			{
				RECT r = {x * (mHeightMap->GetSize().x - 1) / (long)numPatches, 
					y * (mHeightMap->GetSize().y - 1) / (long)numPatches, 
					(x+1) * (mHeightMap->GetSize().x - 1) / (long)numPatches,
					(y+1) * (mHeightMap->GetSize().y - 1) / (long)numPatches};

				TerrainPatch *p = new TerrainPatch();
				p->CreateMesh(this, r, x + y);
				mPatches.push_back(p);
			}
		}
	}
	catch(...)
	{		
	}	
}

void Terrain::RenderPatches( Camera *cam )
{
	mTotalVertices = 0;
	mTotalFaces = 0;
	mTotalPatches = 0;

	for(size_t j = 0; j < mPatches.size(); j++)
	{
		if(  cam->FrustumCubeCheck( mPatches[j]->mBoundingBox.min, mPatches[j]->mBoundingBox.max ) )	
		{
			mPatches[j]->Render(0);
			mTotalVertices += mPatches[j]->mMesh->GetNumVertices();
			mTotalFaces += mPatches[j]->mMesh->GetNumFaces();
			mTotalPatches++;
		}
	}
}

void Terrain::Release()
{
	for(size_t i = 0; i < mPatches.size(); i++)
	{
		if(mPatches[i] != NULL)
		{	
			TerrainPatch *p = mPatches[i];
			p->Release();
			delete p;
		}
	}
	mPatches.clear();

	for(size_t i = 0; i < mBlendMaps.size(); i++)
	{
		if(mBlendMaps[i] != NULL)
			mBlendMaps[i]->Release();
	}
	mBlendMaps.clear();		

	for(size_t i = 0; i < mDiffuseMaps.size(); i++)
	{
		if(mDiffuseMaps[i] != NULL)
			mDiffuseMaps[i]->Release();
	}
	mDiffuseMaps.clear();	

	SAFE_RELEASE( mEffect );
	SAFE_RELEASE( mLightMap );	
	SAFE_RELEASE( mCalculatedBlendMap );	
}

void Terrain::Render( Camera *cam )
{
	if( !mEffect )
		return;

	// Reset stats
	mTotalVertices = 0;
	mTotalFaces = 0;
	mTotalPatches = 0;

	D3DXMATRIX worldMatrix;	

	D3DXVECTOR4 eyePos = *cam->GetEyePt();	

	// Set vertex shader variables
	D3DXMatrixIdentity(&worldMatrix);
	mEffect->SetMatrix("matWorld", &worldMatrix);
	mEffect->SetMatrix("matVP", &(*cam->GetViewMatrix() * *cam->GetProjMatrix()));

	D3DXVECTOR4 v( mDirToLight.x, mDirToLight.y, mDirToLight.z, 1.f);
	mEffect->SetVector("dirToLight", &v);
	mEffect->SetVector("eyePos", &eyePos);

	// Set pixel shader variables
	//mEffect->SetTexture("a", mCalculatedBlendMap);
	mEffect->SetTexture("a", mBlendMaps[0]);
	mEffect->SetTexture("a2", mBlendMaps[1]);
	mEffect->SetTexture("a3", mBlendMaps[2]);
	mEffect->SetTexture("t1", mDiffuseMaps[0]);
	mEffect->SetTexture("t2", mDiffuseMaps[1]);
	mEffect->SetTexture("t3", mDiffuseMaps[2]);
	mEffect->SetTexture("t4", mDiffuseMaps[3]);
	mEffect->SetTexture("t5", mDiffuseMaps[4]);
	mEffect->SetTexture("t6", mDiffuseMaps[5]);
	mEffect->SetTexture("t7", mDiffuseMaps[6]);
	mEffect->SetTexture("t8", mDiffuseMaps[7]);
	mEffect->SetTexture("t9", mDiffuseMaps[8]);
	mEffect->SetTexture("l", mLightMap);

	mEffect->SetTechnique( "TerrainTech" );
	UINT numPasses = 0;
	mEffect->Begin(&numPasses, 0);
	mEffect->BeginPass(0);

	// Render Patches
	for(size_t j = 0; j < mPatches.size(); j++)
	{
		if( cam->FrustumCubeCheck( mPatches[j]->mBoundingBox.min, mPatches[j]->mBoundingBox.max ) )
		{
			mPatches[j]->Render(0);
			mTotalVertices += mPatches[j]->mMesh->GetNumVertices();
			mTotalFaces += mPatches[j]->mMesh->GetNumFaces();
			mTotalPatches++;
		}
	}	

	mEffect->EndPass();
	mEffect->End();	
}

void Terrain::CalculateBlendMap()
{
	// Height ranges used to calculate the color of each channel
	float min_range[] = { 0.0f, 2.f, 6.f };
	float max_range[] = { 2.f, 6.0f, mHeightMap->GetMaxHeight() };

	// Clear old blend map
	if(mCalculatedBlendMap != NULL)
		mCalculatedBlendMap->Release();

	// Create new blend map
	D3DXCreateTexture(mDevice, mHeightMap->GetSize().x, mHeightMap->GetSize().y, 1, D3DUSAGE_DYNAMIC, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, &mCalculatedBlendMap);

	// Lock the texture
	D3DLOCKED_RECT sRect;
	mCalculatedBlendMap->LockRect(0, &sRect, NULL, NULL);
	BYTE *bytes = (BYTE*)sRect.pBits;
	memset( bytes, 0, mHeightMap->GetSize().x*sRect.Pitch );		//Clear texture to black

	// Start filling in the blend map with RGB colors
	for( int i = 0; i < 3; i++ )
	{
		for( int y = 0; y < sRect.Pitch / 4; y++ )
		{
			for( int x = 0; x < sRect.Pitch / 4; x++ )
			{
				int hm_x = mHeightMap->GetSize().x * (x / (float)(sRect.Pitch / 4));
				int hm_y = mHeightMap->GetSize().y * (y / (float)(sRect.Pitch / 4));
				float height = mHeightMap->GetHeightTable()[hm_x + hm_y * mHeightMap->GetSize().x];				

				BYTE *b = bytes + y * sRect.Pitch + x * 4 + i;
				if(height >= min_range[i] && height <= max_range[i])
					*b = 255; // Apply color for this pixel
				else 
					*b = 0;	// No color for this pixel			
			}
		}
	}

	// Unlock the texture
	mCalculatedBlendMap->UnlockRect(0);	
}

void Terrain::CalculateLightMap()
{
	try
	{
		// Clear old light map
		if(mLightMap != NULL)
			mLightMap->Release();

		Point size = mHeightMap->GetSize();
		float xOffset = (float)size.x * 0.5f; 
		float zOffset = (float)size.y * 0.5f;	

		// Create new light map
		D3DXCreateTexture(mDevice, size.x, size.y, 1, D3DUSAGE_DYNAMIC, D3DFMT_L8, D3DPOOL_DEFAULT, &mLightMap);

		// Lock the texture
		D3DLOCKED_RECT sRect;
		mLightMap->LockRect(0, &sRect, NULL, NULL);
		BYTE *bytes = (BYTE*)sRect.pBits;
		memset( bytes, 255, mHeightMap->GetSize().x * sRect.Pitch );		//Clear texture to white

		for(int y = 0; y < sRect.Pitch; y++)
		{
			Progress("Calculating Lightmap", y / (float)sRect.Pitch);

			for(int x = 0; x < sRect.Pitch; x++)
			{
				float terrain_x = (float)size.x * (x / (float)(sRect.Pitch));
				float terrain_z = (float)size.y * (y / (float)(sRect.Pitch));

				// Find patch that the terrain_x, terrain_z is over
				bool done = false;
				for(size_t p = 0; p < mPatches.size() && !done; p++)
				{
					RECT mr = mPatches[p]->mMapRect;

					//Focus within patch maprect or not?
					if(terrain_x >= mr.left && terrain_x < mr.right &&
						terrain_z >= mr.top && terrain_z < mr.bottom)
					{			
						// Collect only the closest intersection
						Ray rayTop(D3DXVECTOR3(terrain_x - xOffset, 10000.0f, terrain_z - zOffset), D3DXVECTOR3( 0.0f, -1.0f, 0.0f));
						float dist = rayTop.Intersect( mPatches[p]->mMesh );

						if(dist >= 0.0f)
						{
							Ray ray(D3DXVECTOR3(terrain_x - xOffset, 10000.0f - dist + 0.01f, terrain_z  - zOffset ), mDirToLight);

							for(size_t p2 = 0; p2 < mPatches.size() && !done; p2++)
							{
								if(ray.Intersect(mPatches[p2]->mBoundingBox) >= 0)
								{
									if(ray.Intersect(mPatches[p2]->mMesh) >= 0) // In shadow
									{
										done = true;
										bytes[y * sRect.Pitch + x] = 128;
									}
								}
							}
							done = true;
						}
					}
				}						
			}
		}

		// Smooth lightmap	
		for(int i = 0; i < 3; i++)
		{
			Progress("Smoothing the Lightmap", i / 3.0f);

			BYTE* tmpBytes = new BYTE[sRect.Pitch * sRect.Pitch];
			memcpy(tmpBytes, sRect.pBits, sRect.Pitch * sRect.Pitch);

			for( int y = 1; y < sRect.Pitch - 1; y++ )
				for( int x = 1; x < sRect.Pitch - 1; x++ )
				{
					long index = y*sRect.Pitch + x;
					BYTE b1 = bytes[index];
					BYTE b2 = bytes[index - 1];
					BYTE b3 = bytes[index - sRect.Pitch];
					BYTE b4 = bytes[index + 1];
					BYTE b5 = bytes[index + sRect.Pitch];

					tmpBytes[index] = (BYTE)((b1 + b2 + b3 + b4 + b5) / 5);
				}

				memcpy(sRect.pBits, tmpBytes, sRect.Pitch * sRect.Pitch);
				delete [] tmpBytes;
		}		
		
		mLightMap->UnlockRect(0);
	}
	catch(...)
	{
		LOG_ERROR( "Error calculating lightmap" );
	}
}